# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
# SPDX-FileCopyrightText: 2023-present Filip Maksimovic <filip.maksimovic@inria.fr>
# SPDX-FileCopyrightText: 2024-present Diego Badillo <diego.badillo@sansano.usm.cl>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Pydantic models used by the controller and server application."""

# pylint: disable=too-few-public-methods,no-name-in-module

from enum import IntEnum
from typing import Any, List, Optional, Union

from pydantic import BaseModel

from dotbot.protocol import ApplicationType, ControlModeType

MAX_POSITION_HISTORY_SIZE = 1000


class DotBotAddressModel(BaseModel):
    """Simple model to hold a DotBot address."""

    address: str


class DotBotCalibrationStateModel(BaseModel):
    """Model that holds the controller LH2 calibration state."""

    state: str


class DotBotCalibrationIndexModel(BaseModel):
    """Model that holds the controller LH2 calibration index."""

    index: int


class MqttPinCodeModel(BaseModel):
    """Pin code used to derive crypto keys for MQTT."""

    pin: int


class DotBotMoveRawCommandModel(BaseModel):
    """Model class that defines a move raw command."""

    left_x: int
    left_y: int
    right_x: int
    right_y: int


class DotBotRgbLedCommandModel(BaseModel):
    """Model class that defines an RGB LED command."""

    red: int
    green: int
    blue: int


class DotBotXGOActionCommandModel(BaseModel):
    """Model class that defines an XGO action command."""

    action: int


class DotBotLH2Position(BaseModel):
    """Position of a DotBot."""

    x: float
    y: float
    z: float


class DotBotControlModeModel(BaseModel):
    """Mode of a DotBot."""

    mode: ControlModeType


class DotBotGPSPosition(BaseModel):
    """GPS position of a DotBot, usually running a SailBot application."""

    latitude: float
    longitude: float


class DotBotWaypoints(BaseModel):
    """Waypoints model."""

    threshold: int
    waypoints: List[Union[DotBotLH2Position, DotBotGPSPosition]]


class DotBotStatus(IntEnum):
    """Status of a DotBot."""

    ALIVE: int = 0
    LOST: int = 1
    DEAD: int = 2


class DotBotQueryModel(BaseModel):
    """Model class used to filter DotBots."""

    max_positions: int = MAX_POSITION_HISTORY_SIZE
    application: Optional[ApplicationType] = None
    mode: Optional[ControlModeType] = None
    status: Optional[DotBotStatus] = None
    swarm: Optional[str] = None


class DotBotNotificationCommand(IntEnum):
    """Notification command of a DotBot."""

    NONE: int = 0
    RELOAD: int = 1
    UPDATE: int = 2
    PIN_CODE_UPDATE: int = 3


class DotBotNotificationUpdate(BaseModel):
    """Update notification model."""

    address: str
    direction: Optional[int]
    wind_angle: Optional[int]
    rudder_angle: Optional[int]
    sail_angle: Optional[int]
    lh2_position: Optional[DotBotLH2Position] = None
    gps_position: Optional[DotBotGPSPosition] = None


class DotBotNotificationModel(BaseModel):
    """Model class used to send controller notifications."""

    cmd: DotBotNotificationCommand
    data: Optional[DotBotNotificationUpdate] = None
    pin_code: Optional[int] = None


class DotBotRequestType(IntEnum):
    """Request received from MQTT client."""

    DOTBOTS: int = 0
    LH2_CALIBRATION_STATE: int = 1


class DotBotRequestModel(BaseModel):
    """Model class used to handle controller request."""

    request: DotBotRequestType
    reply: str


class DotBotReplyModel(BaseModel):
    """Model class used to handle controller replies."""

    request: DotBotRequestType
    data: Any


class DotBotModel(BaseModel):
    """Model class that defines a DotBot."""

    address: str
    application: ApplicationType = ApplicationType.DotBot
    swarm: str = "0000"
    status: DotBotStatus = DotBotStatus.ALIVE
    mode: ControlModeType = ControlModeType.MANUAL
    last_seen: float
    direction: Optional[int] = None
    wind_angle: Optional[int] = None
    rudder_angle: Optional[int] = None
    sail_angle: Optional[int] = None
    move_raw: Optional[DotBotMoveRawCommandModel] = None
    rgb_led: Optional[DotBotRgbLedCommandModel] = None
    lh2_position: Optional[DotBotLH2Position] = None
    gps_position: Optional[DotBotGPSPosition] = None
    waypoints: List[Union[DotBotLH2Position, DotBotGPSPosition]] = []
    waypoints_threshold: int = 40
    position_history: List[Union[DotBotLH2Position, DotBotGPSPosition]] = []



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing a joystick Dotbot controller."""

# pylint: disable=attribute-defined-outside-init

import asyncio
import os
import sys

import click

from dotbot import (
    CONTROLLER_HOSTNAME_DEFAULT,
    CONTROLLER_PORT_DEFAULT,
    DOTBOT_ADDRESS_DEFAULT,
    pydotbot_version,
)
from dotbot.logger import LOGGER, setup_logging
from dotbot.models import DotBotMoveRawCommandModel
from dotbot.protocol import ApplicationType
from dotbot.rest import RestClient

# Pygame support prompt is annoying, it can be hidden using an environment variable
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"
import pygame  # noqa: E402, pylint: disable=wrong-import-order, wrong-import-position

JOYSTICK_HYSTERERIS_THRES = 0.09
JOYSTICK_AXIS_COUNT = 4
REFRESH_PERIOD = 0.05
NULL_POSITION = [0.0, 0.0, 0.0, 0.0]

DOTBOT_APPLICATION_DEFAULT = "sailbot"
APPLICATION_TYPE_MAP = {
    "dotbot": ApplicationType.DotBot,
    "sailbot": ApplicationType.SailBot,
}


class JoystickController:
    """A Dotbot controller for a joystick interface."""

    def __init__(
        self, joystick_index, hostname, port, https, dotbot_address, application
    ):
        """Initialize the joystick controller."""
        self.api = RestClient(hostname, port, https)
        self.dotbots = []
        self.dotbot_address = dotbot_address
        self.application = APPLICATION_TYPE_MAP[application]
        pygame.init()  # pylint: disable=no-member
        pygame.joystick.init()  # joysticks initialization
        self._logger = LOGGER.bind(context=__name__)
        if pygame.joystick.get_count() < joystick_index + 1:
            self._logger.error("No joystick connected")
            sys.exit("Error: No joystick connected.\nExiting program...")
        self.joystick = pygame.joystick.Joystick(
            joystick_index
        )  # instantiation of a joystick
        self.joystick.init()  # initialization of the joystick
        num_axes = self.joystick.get_numaxes()
        if num_axes < JOYSTICK_AXIS_COUNT:
            self._logger.error("Not enough axes")
            sys.exit(
                f"Not enough axes on your joystick. {num_axes} found, expected at least {JOYSTICK_AXIS_COUNT}."
            )
        self.previous_positions = NULL_POSITION
        self._logger.info("Controller initialized", num_axes=num_axes)

    @property
    def active_dotbot(self):
        _active_dotbot = self.dotbot_address
        if _active_dotbot == DOTBOT_ADDRESS_DEFAULT:
            if self.dotbots and self.dotbots[0]["status"] == 0:
                _active_dotbot = self.dotbots[0]["address"]
            else:
                self._logger.info("No active DotBot")
                return
        elif _active_dotbot not in [dotbot["address"] for dotbot in self.dotbots]:
            self._logger.info("Active DotBot not available")
            return
        return _active_dotbot

    def pos_from_joystick(self):
        """Fetch positions of the joystick."""
        pygame.event.pump()  # queue needs to be pumped
        positions = []
        for axis_idx in range(JOYSTICK_AXIS_COUNT):
            axis = self.joystick.get_axis(axis_idx)
            # dead zones
            if axis_idx % 2:
                axis = -axis
            if -JOYSTICK_HYSTERERIS_THRES < axis <= JOYSTICK_HYSTERERIS_THRES:
                axis = 0.0
            # from [-1;1] to [-127;127]
            positions.append(axis * 127)
        return positions

    async def fetch_active_dotbots(self):
        while 1:
            self.dotbots = await self.api.fetch_active_dotbots()
            await asyncio.sleep(1)

    async def start(self):
        """Starts to read continuously joystick positions."""
        asyncio.create_task(self.fetch_active_dotbots())
        while True:
            # fetch positions from joystick
            positions = self.pos_from_joystick()
            if positions != NULL_POSITION or self.previous_positions != NULL_POSITION:
                self._logger.info("refresh positions", positions=positions)
                await self.api.send_move_raw_command(
                    self.active_dotbot,
                    self.application,
                    DotBotMoveRawCommandModel(
                        left_x=int(positions[0]),
                        left_y=int(positions[1]),
                        right_x=int(positions[2]),
                        right_y=int(positions[3]),
                    ),
                )
            self.previous_positions = positions
            await asyncio.sleep(REFRESH_PERIOD)  # 50ms delay between each update


@click.command()
@click.option(
    "-j",
    "--joystick",
    type=int,
    default=0,
    help="Index of the joystick to use. Defaults to 0",
)
@click.option(
    "-h",
    "--hostname",
    type=str,
    default=CONTROLLER_HOSTNAME_DEFAULT,
    help="Hostname of the controller. Defaults to 'localhost'",
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=CONTROLLER_PORT_DEFAULT,
    help=f"HTTP port. Defaults to '{CONTROLLER_PORT_DEFAULT}'",
)
@click.option(
    "-s",
    "--https",
    is_flag=True,
    default=False,
    help="Use HTTPS protocol instead of HTTP",
)
@click.option(
    "-d",
    "--dotbot-address",
    type=str,
    default=DOTBOT_ADDRESS_DEFAULT,
    help=f"Address in hex of the DotBot to control. Defaults to {DOTBOT_ADDRESS_DEFAULT:>0{16}}",
)
@click.option(
    "-a",
    "--application",
    type=click.Choice(["dotbot", "sailbot"]),
    default=DOTBOT_APPLICATION_DEFAULT,
    help=f"Application to control. Defaults to {DOTBOT_APPLICATION_DEFAULT}",
)
@click.option(
    "--log-level",
    type=click.Choice(["debug", "info", "warning", "error"]),
    default="info",
    help="Logging level. Defaults to info",
)
def main(joystick, hostname, port, https, dotbot_address, application, log_level):
    """DotBot joystick controller."""
    print(f"Welcome to the DotBots joystick interface (version: {pydotbot_version()}).")
    setup_logging(None, log_level, ["console"])
    joystick_controller = JoystickController(
        joystick,
        hostname,
        port,
        https,
        dotbot_address,
        application,
    )
    try:
        asyncio.run(joystick_controller.start())
    except (SystemExit, KeyboardInterrupt):
        sys.exit(0)


if __name__ == "__main__":
    main()  # pragma: nocover, pylint: disable=no-value-for-parameter



# SPDX-FileCopyrightText: 2024-present Inria
# SPDX-FileCopyrightText: 2024-present Diego Badillo <diego.badillo@sansano.usm.cl>
# SPDX-FileCopyrightText: 2024-present Mališa Vučinić <malisa.vucinic@inria.fr>
# SPDX-FileCopyrightText: 2024-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Sailbot simulator for the DotBot project."""

import math
import threading
import time
from dataclasses import dataclass
from enum import Enum
from typing import Callable

from numpy import clip

from dotbot import GATEWAY_ADDRESS_DEFAULT, SWARM_ID_DEFAULT
from dotbot.hdlc import hdlc_decode, hdlc_encode
from dotbot.logger import LOGGER
from dotbot.protocol import (
    PROTOCOL_VERSION,
    Advertisement,
    ApplicationType,
    PayloadType,
    ProtocolHeader,
    ProtocolPayload,
    SailBotData,
)

SIM_DELTA_T = 0.01  # second
CONTROL_DELTA_T = 1  # second

RUDDER_ANGLE_MAX = (-math.pi / 6, math.pi / 6)
SAIL_ANGLE_MAX = (0, math.pi / 2)

# constants
EARTH_RADIUS_KM = 6371
ORIGIN_COORD_LAT = 48.825908
ORIGIN_COORD_LON = 2.406433
COS_PHI_0 = 0.658139837


def cartesian2geographical(x, y):
    """Converts cartesian coordinates to geographical coordinates."""

    latitude = (((y * 0.180) / math.pi) / EARTH_RADIUS_KM) + ORIGIN_COORD_LAT
    longitude = (
        ((x * 0.180) / math.pi / COS_PHI_0) / EARTH_RADIUS_KM
    ) + ORIGIN_COORD_LON

    return latitude, longitude


def geographical2cartesian(latitude, longitude):
    """Converts geographical coordinates to cartesian coordinates."""

    x = ((longitude - ORIGIN_COORD_LON) * EARTH_RADIUS_KM * COS_PHI_0 * math.pi) / 0.180
    y = ((latitude - ORIGIN_COORD_LAT) * EARTH_RADIUS_KM * math.pi) / 0.180

    return (x, y)


class SailBotSimulatorMode(Enum):
    """Operation mode of the sailbot simulator."""

    MANUAL = "MANUAL"
    AUTOMATIC = "AUTOMATIC"


@dataclass
class Point:
    """Point class for the sailbot simulator."""

    x: float
    y: float


class SailBotSimulatorLineClass:
    """Simulator zig-zag routine helper class."""

    def __init__(self, point: Point, angle_radians):
        # calculate coefficients for Ax + By + C = 0
        self.A = -math.tan(angle_radians)
        self.B = 1
        self.C = (math.tan(angle_radians) * point.x) - point.y

    def distance2point(self, point: Point):
        """Returns the distance from the point to the line."""
        return abs((self.A * point.x) + (self.B * point.y) + self.C) / math.sqrt(
            self.A**2 + self.B**2
        )

    def line_side(self, point: Point):
        """Am I on the left or the right?."""
        d = (self.A * point.x) + (self.B * point.y) + self.C
        return d >= 0


class SailBotSimulator:
    """Simulator class for the sailbot."""

    def __init__(self, address):
        self.address = address

        self.true_wind_speed = 4  # [m/s]
        self.true_wind_angle = 0.34906  # [rad] (20 degrees)

        # initializations
        self.latitude = 48.832313
        self.longitude = 2.412689
        self.app_wind_angle = 0.0  # [rad]
        self.app_wind_speed = 0.0  # [m/s]

        self.direction = math.pi / 2  # [rad]
        self.x, self.y = geographical2cartesian(self.latitude, self.longitude)
        self.v = 0.0  # speed [m/s]
        self.w = 0.0  # angular velocity [rad/s]

        # inputs received by controller
        self.rudder_slider = 0  # rudder slider
        self.sail_slider = 0  # sail slider

        self.operation_mode: SailBotSimulatorMode = SailBotSimulatorMode.MANUAL

        # autonomous mode initialisations
        self.waypoint_threshold = 0
        self.waypoints = []
        self.next_waypoint = 0
        self.zigzag_flag = False

        self.logger = LOGGER.bind(context=__name__)

    @property
    def header(self):
        return ProtocolHeader(
            destination=int(GATEWAY_ADDRESS_DEFAULT, 16),
            source=int(self.address, 16),
            swarm_id=int(SWARM_ID_DEFAULT, 16),
            application=ApplicationType.SailBot,
            version=PROTOCOL_VERSION,
        )

    def _update_state_space_model(self, rudder_in_rad, sail_length_in_rad):
        # define model parameters
        p1 = 0.03  # drift coefficient [-]
        p2 = 40  # tangential friction [kgs^−1]
        p3 = 6000  # angular friction [kgm]
        p4 = 200  # sail lift [kgs^−1]
        p5 = 1500  # rudder lift [kgs^−1]
        p6 = 0.5  # distance to sail CoE [m]
        p7 = 0.5  # distance to mast [m]
        p8 = 2  # distance to rudder [m]
        p9 = 300  # mass of boat [kg]
        p10 = 400  # moment of inertia [kgm^2]
        p11 = 0.2  # rudder break coefficient [-]

        # get apparent wind speed and angle, f(v,heading,true_wind)
        self._true2apparent_wind()

        # map mainsheet length to actual sail angle
        sail_in_rad = self._mainsheet2sail_angle(
            sail_length_in_rad, self.app_wind_angle
        )

        # rudder and sail forces
        g_r = p5 * self.v**2 * math.sin(rudder_in_rad)
        g_s = p4 * self.app_wind_speed * math.sin(sail_in_rad - self.app_wind_angle)

        # state-space model
        x_dot = self.v * math.cos(
            self.direction
        ) + p1 * self.true_wind_speed * math.cos(self.true_wind_angle)
        y_dot = self.v * math.sin(
            self.direction
        ) + p1 * self.true_wind_speed * math.sin(self.true_wind_angle)
        direction_dot = self.w
        v_dot = (
            g_s * math.sin(sail_in_rad)
            - g_r * p11 * math.sin(rudder_in_rad)
            - p2 * self.v**2
        ) / p9
        w_dot = (
            g_s * (p6 - p7 * math.cos(sail_in_rad))
            - g_r * p8 * math.cos(rudder_in_rad)
            - p3 * self.w * self.v
        ) / p10

        # update state-space variables and apparent wind angle using Euler's method
        self.x += x_dot * SIM_DELTA_T
        self.y += y_dot * SIM_DELTA_T
        self.direction += direction_dot * SIM_DELTA_T
        self.v += v_dot * SIM_DELTA_T
        self.w += w_dot * SIM_DELTA_T

        # get latitude and longitude from cartesian coordinates
        self.latitude, self.longitude = cartesian2geographical(self.x, self.y)

    def _true2apparent_wind(self):
        Wc_aw = (
            self.true_wind_speed * math.cos(self.true_wind_angle - self.direction)
            - self.v,
            self.true_wind_speed * math.sin(self.true_wind_angle - self.direction),
        )

        self.app_wind_speed = math.sqrt(Wc_aw[0] ** 2 + Wc_aw[1] ** 2)
        self.app_wind_angle = math.atan2(Wc_aw[1], Wc_aw[0])

    def _map_slider(self, value, target_min, target_max):
        slider_min, slider_max = -128, 128
        return target_min + (target_max - target_min) * (value - slider_min) / (
            slider_max - slider_min
        )

    def _mainsheet2sail_angle(self, sail_in_length_rad, app_wind_angle):
        if math.cos(app_wind_angle) + math.cos(sail_in_length_rad) > 0:
            # sail is tight
            sail_out_rad = -math.copysign(sail_in_length_rad, math.sin(app_wind_angle))
        else:
            # sail is loose
            sail_out_rad = math.pi + app_wind_angle

        sail_out_rad = (sail_out_rad + math.pi) % (2 * math.pi) - math.pi
        return sail_out_rad

    def simulation_update(self):
        """Updates the state-space model every SIM_DELTA_T seconds convert to radians."""

        rudder_in_rad = self._map_slider(
            self.rudder_slider, RUDDER_ANGLE_MAX[0], RUDDER_ANGLE_MAX[1]
        )
        sail_length_in_rad = self._map_slider(
            self.sail_slider, SAIL_ANGLE_MAX[0], SAIL_ANGLE_MAX[1]
        )

        self._update_state_space_model(rudder_in_rad, sail_length_in_rad)

        return self.encode_serial_output()

    def control_loop_update(self):
        """Control loop for the sailbot in automatic mode."""

        if not self.waypoints:
            return
        self.logger.debug("Loop update")
        logger = self.logger.bind(
            next_waypoint=self.next_waypoint,
            waypoint_threshold=self.waypoint_threshold,
            waypoints=self.waypoints,
        )
        logger.debug("Loop start")

        # update control inputs if in automatic mode
        if self.next_waypoint >= len(self.waypoints):
            logger.debug("Finished!")
            self.waypoints = []
            self.next_waypoint = 0
            self.zigzag_flag = False
            self.operation_mode = SailBotSimulatorMode.MANUAL
            return

        # convert current position and next waypoint to cartesian
        current_x, current_y = geographical2cartesian(self.latitude, self.longitude)
        next_geo = self.waypoints[self.next_waypoint]
        next_x, next_y = geographical2cartesian(next_geo[0], next_geo[1])

        # check if current position is within the threshold
        distance2target = math.sqrt(
            (next_x - current_x) ** 2 + (next_y - current_y) ** 2
        )
        if distance2target < self.waypoint_threshold:
            self.next_waypoint += 1
            self.zigzag_flag = False
            return

        # compute angle between current position and target
        angle2target = math.atan2(next_y - current_y, next_x - current_x)
        angle2target %= 2 * math.pi
        true_wind = self.true_wind_angle % (2 * math.pi)

        # check if to get there, we have to sail against the wind
        in_irons_angle = 0.8  # radians

        upper_bound_irons = (true_wind + math.pi + in_irons_angle) % (2 * math.pi)
        lower_bound_irons = (true_wind + math.pi - in_irons_angle) % (2 * math.pi)

        detected_irons = False

        if lower_bound_irons > upper_bound_irons:
            if angle2target < upper_bound_irons or angle2target > lower_bound_irons:
                detected_irons = True
        else:
            if angle2target < upper_bound_irons and angle2target > lower_bound_irons:
                detected_irons = True

        if detected_irons:
            if self.zigzag_flag is False:
                logger.debug("In irons! Starting zig-zag routine")
                self.zigzag_flag = True

                # initialise zig zag parameters
                self.zigzag_width = 15  # meters
                self.zigzag_dir = False  # where to start zig-zagging (can be improved by choosing the shortest path)
                self.line2target = SailBotSimulatorLineClass(
                    Point(current_x, current_y), angle2target
                )

            # already zig-zagging!
            if (
                self.line2target.distance2point(Point(current_x, current_y))
                > self.zigzag_width
            ):
                self.zigzag_dir = self.line2target.line_side(
                    Point(current_x, current_y)
                )

            # redefine reference heading
            angle2target = upper_bound_irons if self.zigzag_dir else lower_bound_irons

        else:
            # return to normal mode
            self.zigzag_flag = False

        # calculate error between angle2target and current heading
        error_heading2target = angle2target - self.direction
        error_heading2target = (error_heading2target + math.pi) % (
            2 * math.pi
        ) - math.pi

        # map error to rudder angle (proportional controller)
        Kp = -256 / math.pi  # control proportional constant
        self.rudder_slider = int(clip(error_heading2target * Kp, -128, 127))

        # linear map apparent wind angle to sail opening
        app_wind_angle = (self.app_wind_angle + math.pi) % (2 * math.pi) - math.pi
        sail_length = ((SAIL_ANGLE_MAX[0] - SAIL_ANGLE_MAX[1]) / math.pi) * abs(
            app_wind_angle
        ) + SAIL_ANGLE_MAX[1]
        self.sail_slider = int(clip(sail_length * 512 / math.pi - 128, -128, 127))

        logger = self.logger.bind(
            rudder_slider=self.rudder_slider,
            sail_slider=self.sail_slider,
            zigzag_flag=self.zigzag_flag,
        )
        logger.debug("Loop end")

    def decode_serial_input(self, frame):
        """Decode the serial input received from the gateway."""

        payload = ProtocolPayload.from_bytes(hdlc_decode(frame))

        if self.address == hex(payload.header.destination)[2:]:
            if payload.payload_type == PayloadType.CMD_MOVE_RAW:
                self.rudder_slider = (
                    payload.values.left_x - 256
                    if payload.values.left_x > 127
                    else payload.values.left_x
                )
                self.sail_slider = (
                    payload.values.right_y - 256
                    if payload.values.right_y > 127
                    else payload.values.right_y
                )

            if payload.payload_type == PayloadType.GPS_WAYPOINTS:
                self.operation_mode = SailBotSimulatorMode.MANUAL
                self.waypoint_threshold = payload.values.threshold
                self.waypoints = payload.values.waypoints
                if self.waypoints:
                    self.operation_mode = SailBotSimulatorMode.AUTOMATIC

    def encode_serial_output(self):
        """Encode the sailbot data to be sent to the gateway."""

        payload = ProtocolPayload(
            self.header,
            PayloadType.SAILBOT_DATA,
            SailBotData(
                (90 - int(math.degrees(self.direction))) % 360,
                int(self.latitude * 1e6),
                int(self.longitude * 1e6),
                (-int(math.degrees(self.app_wind_angle))) % 360,
                int(
                    math.degrees(
                        self._map_slider(
                            self.rudder_slider, RUDDER_ANGLE_MAX[0], RUDDER_ANGLE_MAX[1]
                        )
                    )
                ),
                int(
                    math.degrees(
                        self._map_slider(
                            self.sail_slider, SAIL_ANGLE_MAX[0], SAIL_ANGLE_MAX[1]
                        )
                    )
                ),
            ),
        )
        return hdlc_encode(payload.to_bytes())

    def advertise(self):
        """Send an adertisement message to the gateway."""
        payload = ProtocolPayload(
            self.header,
            PayloadType.ADVERTISEMENT,
            Advertisement(),
        )
        return hdlc_encode(payload.to_bytes())


class SailBotSimulatorSerialInterface(threading.Thread):
    """Bidirectional serial interface to control simulated robots."""

    def __init__(self, callback: Callable):
        self.sailbots = [
            SailBotSimulator("1234567890123456"),
        ]

        self.callback = callback
        super().__init__(daemon=True)  # automatically close when the main program exits
        self.start()
        self.logger = LOGGER.bind(context=__name__)
        self.logger.info("SailBot simulation started")

    def run(self):
        """Listen continuously at each byte received on the fake serial interface."""
        for sailbot in self.sailbots:
            for byte in sailbot.advertise():
                self.callback(byte.to_bytes(length=1, byteorder="little"))

        next_sim_time = time.time() + SIM_DELTA_T
        next_control_time = time.time() + CONTROL_DELTA_T
        updates = [bytearray()] * len(self.sailbots)
        updates_interval = 0
        while True:
            current_time = time.time()
            # update simulation every SIM_DELTA_T seconds
            if current_time >= next_sim_time:
                for idx, sailbot in enumerate(self.sailbots):
                    updates[idx] = sailbot.simulation_update()
                next_sim_time = current_time + SIM_DELTA_T
            if updates_interval >= 10:
                for update in updates:
                    for byte in update:
                        self.callback(byte.to_bytes(length=1, byteorder="little"))
                        time.sleep(0.001)
                updates_interval = 0
            updates_interval += 1

            # update control inputs every CONTROL_DELTA_T seconds
            if current_time >= next_control_time:
                for sailbot in self.sailbots:
                    if sailbot.operation_mode == SailBotSimulatorMode.AUTOMATIC:
                        sailbot.control_loop_update()

                next_control_time = current_time + CONTROL_DELTA_T
            time.sleep(0.02)

    def write(self, bytes_):
        """Write bytes on the fake serial, similar to the real gateway."""
        for sailbot in self.sailbots:
            sailbot.decode_serial_input(bytes_)



from unittest import mock

import httpx
import pytest

from dotbot.models import DotBotMoveRawCommandModel, DotBotRgbLedCommandModel
from dotbot.protocol import ApplicationType
from dotbot.rest import RestClient


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "response,expected",
    [
        pytest.param(httpx.Response(200, json=[]), [], id="empty"),
        pytest.param(httpx.ConnectError, [], id="http error"),
        pytest.param(httpx.Response(403), [], id="http code error"),
        pytest.param(
            httpx.Response(200, json=[{"address": "test", "status": 1}]),
            [],
            id="none active",
        ),
        pytest.param(
            httpx.Response(200, json=[{"address": "test", "status": 0}]),
            [{"address": "test", "status": 0}],
            id="found",
        ),
    ],
)
@mock.patch("httpx.AsyncClient.get")
async def test_fetch_active_dotbots(get, response, expected):
    if response == httpx.ConnectError:
        get.side_effect = response("error")
    else:
        get.return_value = response
    client = RestClient("localhost", 1234, False)
    result = await client.fetch_active_dotbots()
    get.assert_called_once()
    assert result == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "response,application,command",
    [
        pytest.param(
            httpx.Response(200),
            ApplicationType.DotBot,
            DotBotMoveRawCommandModel(left_x=0, left_y=0, right_x=0, right_y=0),
            id="dotbot",
        ),
        pytest.param(
            httpx.Response(200),
            ApplicationType.SailBot,
            DotBotMoveRawCommandModel(left_x=0, left_y=0, right_x=0, right_y=0),
            id="sailbot",
        ),
        pytest.param(
            httpx.ConnectError,
            ApplicationType.DotBot,
            DotBotMoveRawCommandModel(left_x=0, left_y=0, right_x=0, right_y=0),
            id="http error",
        ),
        pytest.param(
            httpx.Response(403),
            ApplicationType.DotBot,
            DotBotMoveRawCommandModel(left_x=0, left_y=0, right_x=0, right_y=0),
            id="invalid http code",
        ),
    ],
)
@mock.patch("httpx.AsyncClient.put")
async def test_send_move_raw_command(put, response, application, command):
    if response == httpx.ConnectError:
        put.side_effect = response("error")
    else:
        put.return_value = response
    client = RestClient("localhost", 1234, False)
    await client.send_move_raw_command("test", application, command)
    put.assert_called_once()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "response,command",
    [
        pytest.param(
            httpx.Response(200),
            DotBotRgbLedCommandModel(red=0, green=0, blue=0),
            id="ok",
        ),
        pytest.param(
            httpx.ConnectError,
            DotBotRgbLedCommandModel(red=0, green=0, blue=0),
            id="http error",
        ),
        pytest.param(
            httpx.Response(403),
            DotBotRgbLedCommandModel(red=0, green=0, blue=0),
            id="invalid http code",
        ),
    ],
)
@mock.patch("httpx.AsyncClient.put")
async def test_send_rgb_led_command(put, response, command):
    if response == httpx.ConnectError:
        put.side_effect = response("error")
    else:
        put.return_value = response
    client = RestClient("localhost", 1234, False)
    await client.send_rgb_led_command("test", command)
    put.assert_called_once()



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing a keyboard Dotbot controller."""

import asyncio
import sys
from dataclasses import dataclass
from enum import Enum

import click

try:
    from pynput import keyboard
except ImportError:
    # On the CI, pynput cannot be imported because of missing X server. Mock
    # the pynput keyboard instead
    from unittest import mock

    keyboard = mock.MagicMock()

from dotbot import (
    CONTROLLER_HOSTNAME_DEFAULT,
    CONTROLLER_PORT_DEFAULT,
    DOTBOT_ADDRESS_DEFAULT,
    pydotbot_version,
)
from dotbot.logger import LOGGER, setup_logging
from dotbot.models import DotBotMoveRawCommandModel, DotBotRgbLedCommandModel
from dotbot.protocol import ApplicationType
from dotbot.rest import RestClient

DOTBOT_APPLICATION_DEFAULT = "dotbot"
APPLICATION_TYPE_MAP = {
    "dotbot": ApplicationType.DotBot,
    "sailbot": ApplicationType.SailBot,
}
DIR_KEYS = [
    keyboard.Key.up,
    keyboard.Key.down,
    keyboard.Key.left,
    keyboard.Key.right,
]
COLOR_KEYS = ["r", "g", "b", "y", "p", "w", "n"]


class MotorSpeeds(Enum):
    """Levels used for motor speeds."""

    NORMAL = 84
    BOOST = 100
    SUPERBOOST = 127


def rgb_from_key(key):
    """Compute the RGB values from a key.

    >>> rgb_from_key("r")
    [255, 0, 0]
    >>> rgb_from_key("g")
    [0, 255, 0]
    >>> rgb_from_key("b")
    [0, 0, 255]
    >>> rgb_from_key("y")
    [255, 255, 0]
    >>> rgb_from_key("p")
    [255, 0, 255]
    >>> rgb_from_key("w")
    [255, 255, 255]
    >>> rgb_from_key("n")
    [0, 0, 0]
    >>> rgb_from_key("a")
    [0, 0, 0]
    >>> rgb_from_key("-")
    [0, 0, 0]
    """
    if key == "r":
        result = [255, 0, 0]
    elif key == "g":
        result = [0, 255, 0]
    elif key == "b":
        result = [0, 0, 255]
    elif key == "y":
        result = [255, 255, 0]
    elif key == "p":
        result = [255, 0, 255]
    elif key == "w":
        result = [255, 255, 255]
    else:  # n
        result = [0, 0, 0]
    return result


class KeyboardEventType(Enum):
    """Supported types of keyboard events."""

    PRESSED = 0
    RELEASED = 1


@dataclass
class KeyboardEvent:
    """Data class that handles data of a keyboard event."""

    type_: KeyboardEventType
    key: keyboard.Key


class KeyboardController:
    """Dotbot controller for a keyboard interface."""

    def __init__(self, hostname, port, https, dotbot_address, application):
        """Initializes the keyboard controller."""
        self.api = RestClient(hostname, port, https)
        self.dotbots = []
        self.dotbot_address = dotbot_address
        self.application = APPLICATION_TYPE_MAP[application]
        self.previous_speeds = (0, 0)
        self.active_keys = []
        self.event_queue = asyncio.Queue()
        self._logger = LOGGER.bind(context=__name__)
        self._logger.info("Controller initialized")

    @property
    def active_dotbot(self):
        _active_dotbot = self.dotbot_address
        if _active_dotbot == DOTBOT_ADDRESS_DEFAULT:
            if self.dotbots and self.dotbots[0]["status"] == 0:
                _active_dotbot = self.dotbots[0]["address"]
            else:
                self._logger.info("No active DotBot")
                return
        elif _active_dotbot not in [dotbot["address"] for dotbot in self.dotbots]:
            self._logger.info("Active DotBot not available")
            return
        return _active_dotbot

    async def update_active_keys(self):
        """Coroutine used to handle keyboard events asynchronously."""
        event_loop = asyncio.get_event_loop()

        def on_press(key):
            """Callback called on each keyboard key press event."""
            if key in self.active_keys:
                return
            event_loop.call_soon_threadsafe(
                self.event_queue.put_nowait,
                KeyboardEvent(KeyboardEventType.PRESSED, key),
            )
            self._logger.debug("key pressed", key=key)

        def on_release(key):
            """Callback called on each keyboard key release event."""
            if key not in self.active_keys:
                return
            event_loop.call_soon_threadsafe(
                self.event_queue.put_nowait,
                KeyboardEvent(KeyboardEventType.RELEASED, key),
            )
            self._logger.debug("key released", key=key)

        listener = keyboard.Listener(on_press=on_press, on_release=on_release)
        listener.start()

        while 1:
            event = await self.event_queue.get()
            if event.type_ == KeyboardEventType.RELEASED:
                while event.key in self.active_keys:
                    self.active_keys.remove(event.key)
            if event.type_ == KeyboardEventType.PRESSED:
                if hasattr(event.key, "char") and event.key.char in COLOR_KEYS:
                    red, green, blue = rgb_from_key(event.key.char)
                    self._logger.info("color pressed", red=red, green=green, blue=blue)
                    await self.api.send_rgb_led_command(
                        self.active_dotbot,
                        DotBotRgbLedCommandModel(red=red, green=green, blue=blue),
                    )
                if event.key not in self.active_keys:
                    self.active_keys.append(event.key)

    def speeds_from_keys(self):  # pylint: disable=too-many-return-statements
        """Computes the left/right wheels speeds from current key pressed."""
        if any(key in self.active_keys for key in DIR_KEYS):
            speed = MotorSpeeds.NORMAL
            if keyboard.Key.ctrl in self.active_keys:
                speed = MotorSpeeds.BOOST
                if keyboard.Key.alt in self.active_keys:
                    speed = MotorSpeeds.SUPERBOOST
            if (
                keyboard.Key.up in self.active_keys
                and keyboard.Key.left in self.active_keys
            ):
                return speed.value * 0.75, speed.value
            if (
                keyboard.Key.up in self.active_keys
                and keyboard.Key.right in self.active_keys
            ):
                return speed.value, speed.value * 0.75
            if (
                keyboard.Key.down in self.active_keys
                and keyboard.Key.left in self.active_keys
            ):
                return -speed.value * 0.75, -speed.value
            if (
                keyboard.Key.down in self.active_keys
                and keyboard.Key.right in self.active_keys
            ):
                return -speed.value, -speed.value * 0.75
            if keyboard.Key.up in self.active_keys:
                return speed.value, speed.value
            if keyboard.Key.down in self.active_keys:
                return -speed.value, -speed.value
            if keyboard.Key.left in self.active_keys:
                return 0, speed.value
            if keyboard.Key.right in self.active_keys:
                return speed.value, 0
        return 0, 0

    async def refresh_speeds(self):
        """Refresh the motor speeds and send an update if needed."""
        left_speed, right_speed = self.speeds_from_keys()
        if (left_speed, right_speed) != (0, 0) or self.previous_speeds != (0, 0):
            self._logger.info("refresh speeds", left=left_speed, right=right_speed)
            await self.api.send_move_raw_command(
                self.active_dotbot,
                self.application,
                DotBotMoveRawCommandModel(
                    left_x=0, left_y=left_speed, right_x=0, right_y=right_speed
                ),
            )
        self.previous_speeds = (left_speed, right_speed)
        await asyncio.sleep(0.05)

    async def fetch_active_dotbots(self):
        while 1:
            self.dotbots = await self.api.fetch_active_dotbots()
            await asyncio.sleep(1)

    async def start(self):
        """Starts to continuously listen on keyboard key press/release events."""
        asyncio.create_task(self.fetch_active_dotbots())
        asyncio.create_task(self.update_active_keys())
        while 1:
            await self.refresh_speeds()


@click.command()
@click.option(
    "-h",
    "--hostname",
    type=str,
    default=CONTROLLER_HOSTNAME_DEFAULT,
    help="Hostname of the controller. Defaults to 'localhost'",
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=CONTROLLER_PORT_DEFAULT,
    help=f"HTTP port. Defaults to '{CONTROLLER_PORT_DEFAULT}'",
)
@click.option(
    "-s",
    "--https",
    is_flag=True,
    default=False,
    help="Use HTTPS protocol instead of HTTP",
)
@click.option(
    "-d",
    "--dotbot-address",
    type=str,
    default=DOTBOT_ADDRESS_DEFAULT,
    help=f"Address in hex of the DotBot to control. Defaults to {DOTBOT_ADDRESS_DEFAULT:>0{16}}",
)
@click.option(
    "-a",
    "--application",
    type=click.Choice(["dotbot", "sailbot"]),
    default=DOTBOT_APPLICATION_DEFAULT,
    help=f"Application to control. Defaults to {DOTBOT_APPLICATION_DEFAULT}",
)
@click.option(
    "--log-level",
    type=click.Choice(["debug", "info", "warning", "error"]),
    default="info",
    help="Logging level. Defaults to info",
)
def main(hostname, port, https, dotbot_address, application, log_level):
    """DotBot keyboard controller."""
    print(f"Welcome to the DotBots keyboard interface (version: {pydotbot_version()}).")
    setup_logging(None, log_level, ["console"])
    keyboard_controller = KeyboardController(
        hostname,
        port,
        https,
        dotbot_address,
        application,
    )
    try:
        asyncio.run(keyboard_controller.start())
    except (SystemExit, KeyboardInterrupt):
        sys.exit(0)


if __name__ == "__main__":
    main()  # pragma: nocover, pylint: disable=no-value-for-parameter



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module containing client code to interact with the controller REST API."""

import httpx

from dotbot.logger import LOGGER
from dotbot.models import DotBotStatus
from dotbot.protocol import ApplicationType


class RestClient:
    """Client to interact with the controller REST API."""

    def __init__(self, hostname, port, https):
        self.hostname = hostname
        self.port = port
        self.https = https
        self._logger = LOGGER.bind(context=__name__)

    @property
    def base_url(self):
        """Returns the base URL of the controller REST API."""
        return f"{'https' if self.https else 'http'}://{self.hostname}:{self.port}/controller"

    async def fetch_active_dotbots(self):
        """Fetch active DotBots."""
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(
                    f"{self.base_url}/dotbots",
                    headers={
                        "Accept": "application/json",
                    },
                )
            except httpx.ConnectError as exc:
                self._logger.warning(f"Failed to fetch dotbots: {exc}")
            else:
                if response.status_code != 200:
                    self._logger.warning(
                        f"Failed to fetch dotbots: {response} {response.text}"
                    )
                else:
                    return [
                        dotbot
                        for dotbot in response.json()
                        if dotbot["status"] == DotBotStatus.ALIVE.value
                    ]
        return []

    async def _send_command(self, address, application, resource, command):
        async with httpx.AsyncClient() as client:
            try:
                response = await client.put(
                    f"{self.base_url}/dotbots"
                    f"/{address}/{application.value}/{resource}",
                    headers={
                        "Accept": "application/json",
                        "Content-Type": "application/json",
                    },
                    content=command.model_dump_json(),
                )
            except httpx.ConnectError as exc:
                self._logger.warning(f"Failed to send command: {exc}")
                return
            if response.status_code != 200:
                self._logger.error(
                    "Cannot send command",
                    response=str(response),
                    status_code=response.status_code,
                    content=str(response.text),
                )

    async def send_move_raw_command(self, address, application, command):
        """Send a move raw command to a DotBot."""
        await self._send_command(address, application, "move_raw", command)

    async def send_rgb_led_command(self, address, command):
        """Send an RGB LED command to a DotBot."""
        await self._send_command(address, ApplicationType.SailBot, "rgb_led", command)



"""Test module for controller base class."""

import asyncio
import time
from dataclasses import dataclass
from typing import List
from unittest.mock import patch

import pytest
import serial

from dotbot.controller import Controller, ControllerSettings, gps_distance, lh2_distance
from dotbot.hdlc import hdlc_encode
from dotbot.models import DotBotGPSPosition, DotBotLH2Position, DotBotModel
from dotbot.protocol import (
    PayloadType,
    ProtocolData,
    ProtocolField,
    ProtocolHeader,
    ProtocolPayload,
)


@dataclass
class ProtocolDataTest(ProtocolData):
    test: int = 0x1234

    @property
    def fields(self) -> List[ProtocolField]:
        """Returns the list of fields in this data."""
        return [
            ProtocolField(ProtocolDataTest.test, "test", 2),
        ]

    @staticmethod
    def from_bytes(bytes_: bytes) -> ProtocolData:
        """Returns a ProtocolData instance from a bytearray."""
        return ProtocolDataTest(bytes_[0:2])


@pytest.mark.asyncio
@patch("dotbot.serial_interface.serial.Serial.write")
@patch("dotbot.serial_interface.serial.Serial.open")
@patch("dotbot.serial_interface.serial.Serial.flush")
async def test_controller(_, __, serial_write, capsys):
    """Check controller subclass instanciation and write to serial."""
    settings = ControllerSettings("/dev/null", "115200", "0", "456", "78")
    controller = Controller(settings)
    controller.dotbots.update(
        {
            "0000000000000000": DotBotModel(
                address="0000000000000000", last_seen=time.time()
            )
        }
    )
    controller.serial = serial.Serial(settings.port, settings.baudrate)
    payload = ProtocolPayload(
        ProtocolHeader(0, 0, 0, 0, 0),
        PayloadType.CMD_MOVE_RAW,
        ProtocolDataTest(),
    )
    # smoke test for the from_bytes static method of ProtocolDataTest
    assert len(ProtocolDataTest.from_bytes(bytearray([1, 2])).fields) == 1
    controller.send_payload(payload)
    assert serial_write.call_count == 1
    payload_expected = hdlc_encode(payload.to_bytes())
    assert serial_write.call_args_list[0].args[0] == payload_expected


@pytest.mark.asyncio
@patch("dotbot.serial_interface.serial.Serial.write")
@patch("dotbot.serial_interface.serial.Serial.open")
@patch("dotbot.serial_interface.serial.Serial.flush")
async def test_controller_dont_send(_, __, serial_write):
    """Check controller subclass instanciation and write to serial."""
    settings = ControllerSettings("/dev/null", "115200", "0", "456", "78")
    controller = Controller(settings)
    dotbot = DotBotModel(address="0000000000000000", last_seen=time.time())
    controller.dotbots.update({dotbot.address: dotbot})
    controller.serial = serial.Serial(settings.port, settings.baudrate)
    payload = ProtocolPayload(
        ProtocolHeader(123, 0, 0, 0, 0),
        PayloadType.CMD_MOVE_RAW,
        ProtocolDataTest(),
    )
    # DotBot is not in the controller known dotbot, so the payload won't be sent
    controller.send_payload(payload)
    assert serial_write.call_count == 0


def test_controller_saibot_simulator():
    """Check controller called for sailbot simulator."""

    async def start_simulator():
        settings = ControllerSettings(
            "sailbot-simulator", "115200", "0", "456", "78", 8000
        )
        controller = Controller(settings)
        try:
            await asyncio.wait_for(controller.run(), timeout=0.5)
        except asyncio.TimeoutError:
            pass

    loop = asyncio.get_event_loop()
    loop.run_until_complete(start_simulator())


@patch("dotbot.controller.QrkeyController.start")
def test_controller_dotbot_simulator(_):
    """Check controller called for dotbot simulator."""

    async def start_simulator():
        settings = ControllerSettings(
            "dotbot-simulator", "115200", "0", "456", "78", 8001
        )
        controller = Controller(settings)
        try:
            await asyncio.wait_for(controller.run(), timeout=0.5)
        except asyncio.TimeoutError:
            pass

    loop = asyncio.get_event_loop()
    loop.run_until_complete(start_simulator())


@pytest.mark.parametrize(
    "last,new,result",
    [
        (DotBotLH2Position(x=0, y=0, z=0), DotBotLH2Position(x=0, y=0, z=0), 0.0),
        (DotBotLH2Position(x=1, y=0, z=0), DotBotLH2Position(x=0, y=0, z=0), 1.0),
        (DotBotLH2Position(x=0, y=1, z=0), DotBotLH2Position(x=0, y=0, z=0), 1.0),
    ],
)
def test_lh2_distance(last, new, result):
    assert lh2_distance(last, new) == result


@pytest.mark.parametrize(
    "last,new,result",
    [
        (
            DotBotGPSPosition(latitude=45.7597, longitude=4.8422),  # Lyon
            DotBotGPSPosition(latitude=48.8567, longitude=2.3508),  # Paris
            392217.25594,
        ),
        (
            DotBotGPSPosition(latitude=51.509865, longitude=-0.118092),  # London
            DotBotGPSPosition(latitude=48.8567, longitude=2.3508),  # Paris
            343374.55271,
        ),
    ],
)
def test_gps_distance(last, new, result):
    assert gps_distance(last, new) == pytest.approx(result)



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
# SPDX-FileCopyrightText: 2023-present Filip Maksimovic <filip.maksimovic@inria.fr>
# SPDX-FileCopyrightText: 2024-present Diego Badillo <diego.badillo@sansano.usm.cl>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Interface of the Dotbot controller."""

import asyncio
import json
import math
import time
import webbrowser
from binascii import hexlify
from dataclasses import dataclass
from typing import Dict, List, Optional

import serial
import uvicorn
import websockets
from fastapi import WebSocket
from haversine import Unit, haversine
from pydantic import ValidationError
from pydantic.tools import parse_obj_as
from qrkey import QrkeyController, SubscriptionModel

from dotbot import (
    CONTROLLER_PORT_DEFAULT,
    DOTBOT_ADDRESS_DEFAULT,
    GATEWAY_ADDRESS_DEFAULT,
)
from dotbot.dotbot_simulator import DotBotSimulatorSerialInterface
from dotbot.hdlc import HDLCHandler, HDLCState, hdlc_encode
from dotbot.lighthouse2 import LighthouseManager, LighthouseManagerState
from dotbot.logger import LOGGER
from dotbot.models import (
    MAX_POSITION_HISTORY_SIZE,
    DotBotCalibrationIndexModel,
    DotBotGPSPosition,
    DotBotLH2Position,
    DotBotModel,
    DotBotMoveRawCommandModel,
    DotBotNotificationCommand,
    DotBotNotificationModel,
    DotBotNotificationUpdate,
    DotBotQueryModel,
    DotBotReplyModel,
    DotBotRequestModel,
    DotBotRequestType,
    DotBotRgbLedCommandModel,
    DotBotStatus,
    DotBotWaypoints,
    DotBotXGOActionCommandModel,
)
from dotbot.protocol import (
    PROTOCOL_VERSION,
    ApplicationType,
    CommandMoveRaw,
    CommandRgbLed,
    CommandXgoAction,
    GPSPosition,
    GPSWaypoints,
    LH2Location,
    LH2Waypoints,
    PayloadType,
    ProtocolHeader,
    ProtocolPayload,
    ProtocolPayloadParserException,
)
from dotbot.sailbot_simulator import SailBotSimulatorSerialInterface
from dotbot.serial_interface import SerialInterface, SerialInterfaceException
from dotbot.server import api

# from dotbot.models import (
#     DotBotModel,
#     DotBotGPSPosition,
#     DotBotLH2Position,
#     DotBotRgbLedCommandModel,
# )


CONTROLLERS = {}
LOST_DELAY = 5  # seconds
DEAD_DELAY = 60  # seconds
LH2_POSITION_DISTANCE_THRESHOLD = 0.01
GPS_POSITION_DISTANCE_THRESHOLD = 5  # meters


class ControllerException(Exception):
    """Exception raised by Dotbot controllers."""


@dataclass
class ControllerSettings:
    """Data class that holds controller settings."""

    port: str
    baudrate: int
    dotbot_address: str
    gw_address: str
    swarm_id: str
    controller_port: int = CONTROLLER_PORT_DEFAULT
    webbrowser: bool = False
    handshake: bool = False
    verbose: bool = False


def lh2_distance(last: DotBotLH2Position, new: DotBotLH2Position) -> float:
    """Helper function that computes the distance between 2 LH2 positions."""
    return math.sqrt(((new.x - last.x) ** 2) + ((new.y - last.y) ** 2))


def gps_distance(last: DotBotGPSPosition, new: DotBotGPSPosition) -> float:
    """Helper function that computes the distance between 2 GPS positions in m."""
    return haversine(
        (last.latitude, last.longitude), (new.latitude, new.longitude), unit=Unit.METERS
    )


class Controller:
    """Abstract base class of specific implementations of Dotbot controllers."""

    def __init__(self, settings: ControllerSettings):
        self.dotbots: Dict[str, DotBotModel] = {}
        # self.dotbots: Dict[str, DotBotModel] = {
        #     "0000000000000001": DotBotModel(
        #         address="0000000000000001",
        #         last_seen=time.time(),
        #         lh2_position=DotBotLH2Position(x=0.5, y=0.5, z=0),
        #         rgb_led=DotBotRgbLedCommandModel(red=255, green=0, blue=0),
        #     ),
        #     "0000000000000002": DotBotModel(
        #         address="0000000000000002",
        #         last_seen=time.time(),
        #         lh2_position=DotBotLH2Position(x=0.2, y=0.2, z=0),
        #         rgb_led=DotBotRgbLedCommandModel(red=0, green=255, blue=0),
        #     ),
        #     "0000000000000003": DotBotModel(
        #         address="0000000000000003",
        #         last_seen=time.time(),
        #     ),
        #     "0000000000000004": DotBotModel(
        #         address="0000000000000004",
        #         application=ApplicationType.SailBot,
        #         last_seen=time.time(),
        #         wind_angle=135,
        #         rotation=49,
        #         gps_position=DotBotGPSPosition(latitude=48.832313766146896, longitude=2.4126897594949184),
        #     ),
        # }
        self.logger = LOGGER.bind(context=__name__)
        self.header = ProtocolHeader(
            destination=int(DOTBOT_ADDRESS_DEFAULT, 16),
            source=int(settings.gw_address, 16),
            swarm_id=int(settings.swarm_id, 16),
            application=ApplicationType.DotBot,
            version=PROTOCOL_VERSION,
        )
        self.settings = settings
        self.hdlc_handler = HDLCHandler()
        self.serial = None
        self.websockets = []
        self.lh2_manager = LighthouseManager()
        self.api = api
        api.controller = self
        self.qrkey = QrkeyController(self.on_request, LOGGER, root_topic="/pydotbot")
        self.subscriptions = [
            SubscriptionModel(
                topic="/command/+/+/+/move_raw", callback=self.on_command_move_raw
            ),
            SubscriptionModel(
                topic="/command/+/+/+/rgb_led", callback=self.on_command_rgb_led
            ),
            SubscriptionModel(
                topic="/command/+/+/+/xgo_action", callback=self.on_command_xgo_action
            ),
            SubscriptionModel(
                topic="/command/+/+/+/waypoints", callback=self.on_command_waypoints
            ),
            SubscriptionModel(
                topic="/command/+/+/+/clear_position_history",
                callback=self.on_command_clear_position_history,
            ),
            SubscriptionModel(topic="/lh2/add", callback=self.on_lh2_add),
            SubscriptionModel(topic="/lh2/start", callback=self.on_lh2_start),
        ]

    def on_command_move_raw(self, topic, payload):
        """Called when a move raw command is received."""
        logger = self.logger.bind(command="move_raw", topic=topic)
        topic_split = topic.split("/")[2:]
        if len(topic_split) != 4 or topic_split[-1] != "move_raw":
            logger.warning("Invalid move_raw command topic")
            return
        swarm_id, address, application, _ = topic_split
        try:
            command = DotBotMoveRawCommandModel(**payload)
        except ValidationError as exc:
            self.logger.warning(f"Invalid move raw command: {exc.errors()}")
            return
        logger.bind(
            address=address,
            application=ApplicationType(int(application)).name,
            **command.model_dump(),
        )
        if address not in self.dotbots:
            logger.warning("DotBot not found")
            return
        logger.info("Sending command")
        header = ProtocolHeader(
            destination=int(address, 16),
            source=int(self.settings.gw_address, 16),
            swarm_id=int(swarm_id, 16),
            application=ApplicationType(int(application)),
            version=PROTOCOL_VERSION,
        )
        payload = ProtocolPayload(
            header,
            PayloadType.CMD_MOVE_RAW,
            CommandMoveRaw(
                left_x=command.left_x,
                left_y=command.left_y,
                right_x=command.right_x,
                right_y=command.right_y,
            ),
        )
        self.send_payload(payload)
        self.dotbots[address].move_raw = command

    def on_command_rgb_led(self, topic, payload):
        """Called when an rgb led command is received."""
        logger = self.logger.bind(command="rgb_led", topic=topic)
        topic_split = topic.split("/")[2:]
        if len(topic_split) != 4 or topic_split[-1] != "rgb_led":
            logger.warning("Invalid rgb_led command topic")
            return
        swarm_id, address, application, _ = topic_split
        try:
            command = DotBotRgbLedCommandModel(**payload)
        except ValidationError as exc:
            LOGGER.warning(f"Invalid rgb led command: {exc.errors()}")
            return
        logger = logger.bind(
            address=address,
            application=ApplicationType(int(application)).name,
            **command.model_dump(),
        )
        if address not in self.dotbots:
            logger.warning("DotBot not found")
            return
        logger.info("Sending command")
        header = ProtocolHeader(
            destination=int(address, 16),
            source=int(self.settings.gw_address, 16),
            swarm_id=int(swarm_id, 16),
            application=ApplicationType(int(application)),
            version=PROTOCOL_VERSION,
        )
        payload = ProtocolPayload(
            header,
            PayloadType.CMD_RGB_LED,
            CommandRgbLed(command.red, command.green, command.blue),
        )
        self.send_payload(payload)
        self.dotbots[address].rgb_led = command
        self.qrkey.publish(
            "/notify",
            DotBotNotificationModel(cmd=DotBotNotificationCommand.RELOAD).model_dump(
                exclude_none=True
            ),
        )

    def on_command_xgo_action(self, topic, payload):
        """Called when an rgb led command is received."""
        logger = self.logger.bind(command="xgo_action", topic=topic)
        topic_split = topic.split("/")[2:]
        if len(topic_split) != 4 or topic_split[-1] != "xgo_action":
            logger.warning("Invalid xgo_action command topic")
            return
        swarm_id, address, application, _ = topic_split
        try:
            command = DotBotXGOActionCommandModel(**payload)
        except ValidationError as exc:
            LOGGER.warning(f"Invalid rgb led command: {exc.errors()}")
            return
        logger = logger.bind(
            address=address,
            application=ApplicationType(int(application)).name,
            **command.model_dump(),
        )
        if address not in self.dotbots:
            logger.warning("DotBot not found")
            return
        logger.info("Sending command")
        header = ProtocolHeader(
            destination=int(address, 16),
            source=int(self.settings.gw_address, 16),
            swarm_id=int(swarm_id, 16),
            application=ApplicationType(int(application)),
            version=PROTOCOL_VERSION,
        )
        payload = ProtocolPayload(
            header,
            PayloadType.CMD_XGO_ACTION,
            CommandXgoAction(command.action),
        )
        self.send_payload(payload)

    def on_command_waypoints(self, topic, payload):
        """Called when a list of waypoints is received."""
        logger = self.logger.bind(command="waypoints", topic=topic)
        topic_split = topic.split("/")[2:]
        if len(topic_split) != 4 or topic_split[-1] != "waypoints":
            logger.warning("Invalid waypoints command topic")
            return
        swarm_id, address, application, _ = topic_split
        command = parse_obj_as(DotBotWaypoints, payload)
        logger = logger.bind(
            address=address,
            application=ApplicationType(int(application)).name,
            threshold=command.threshold,
            length=len(command.waypoints),
        )
        if address not in self.dotbots:
            logger.warning("DotBot not found")
            return
        logger.info("Sending command")
        header = ProtocolHeader(
            destination=int(address, 16),
            source=int(self.settings.gw_address, 16),
            swarm_id=int(swarm_id, 16),
            application=ApplicationType(int(application)),
            version=PROTOCOL_VERSION,
        )
        waypoints_list = command.waypoints
        if ApplicationType(int(application)) == ApplicationType.SailBot:
            if self.dotbots[address].gps_position is not None:
                waypoints_list = [
                    self.dotbots[address].gps_position
                ] + command.waypoints
            payload = ProtocolPayload(
                header,
                PayloadType.GPS_WAYPOINTS,
                GPSWaypoints(
                    threshold=command.threshold,
                    waypoints=[
                        GPSPosition(
                            latitude=int(waypoint.latitude * 1e6),
                            longitude=int(waypoint.longitude * 1e6),
                        )
                        for waypoint in command.waypoints
                    ],
                ),
            )
        else:  # DotBot application
            if self.dotbots[address].lh2_position is not None:
                waypoints_list = [
                    self.dotbots[address].lh2_position
                ] + command.waypoints
            payload = ProtocolPayload(
                header,
                PayloadType.LH2_WAYPOINTS,
                LH2Waypoints(
                    threshold=command.threshold,
                    waypoints=[
                        LH2Location(
                            pos_x=int(waypoint.x * 1e6),
                            pos_y=int(waypoint.y * 1e6),
                            pos_z=int(waypoint.z * 1e6),
                        )
                        for waypoint in command.waypoints
                    ],
                ),
            )
        self.send_payload(payload)
        self.dotbots[address].waypoints = waypoints_list
        self.dotbots[address].waypoints_threshold = command.threshold
        self.qrkey.publish(
            "/notify",
            DotBotNotificationModel(cmd=DotBotNotificationCommand.RELOAD).model_dump(
                exclude_none=True
            ),
        )

    def on_command_clear_position_history(self, topic, _):
        """Called when a clear position history command is received."""
        logger = self.logger.bind(command="clear_position_history", topic=topic)
        topic_split = topic.split("/")[2:]
        if len(topic_split) != 4 or topic_split[-1] != "clear_position_history":
            logger.warning("Invalid clear_position_history command topic")
            return
        _, address, application, _ = topic_split
        logger = logger.bind(
            address=address,
            application=ApplicationType(int(application)).name,
        )
        if address not in self.dotbots:
            logger.warning("DotBot not found")
            return
        logger.info("Sending command")
        self.dotbots[address].position_history = []
        self.qrkey.publish(
            "/notify",
            DotBotNotificationModel(cmd=DotBotNotificationCommand.RELOAD).model_dump(
                exclude_none=True
            ),
        )

    def on_lh2_add(self, topic, payload):
        """Called when an lh2 calibration point is added."""
        logger = self.logger.bind(lh2="add", topic=topic)
        topic_split = topic.split("/")[1:]
        if len(topic_split) != 2 or topic_split[-1] != "add":
            logger.warning("Invalid lh2 add topic")
            return
        try:
            payload = DotBotCalibrationIndexModel(**payload)
        except ValidationError as exc:
            self.logger.warning(f"Invalid calibration index payload: {exc.errors()}")
            return
        logger = self.logger.bind(**payload.model_dump())
        logger.info("Add calibration point")
        self.lh2_manager.add_calibration_point(payload.index)

    def on_lh2_start(self, topic, _):
        """Called to start the lh2 calibration."""
        logger = self.logger.bind(lh2="start", topic=topic)
        topic_split = topic.split("/")[1:]
        if len(topic_split) != 2 or topic_split[-1] != "start":
            logger.warning("Invalid lh2 start topic")
            return
        logger.info("Start calibration")
        self.lh2_manager.compute_calibration()

    def on_request(self, payload):
        logger = LOGGER.bind(topic="/request")
        logger.info("Request received")
        try:
            request = DotBotRequestModel(**payload)
        except ValidationError as exc:
            logger.warning(f"Invalid request: {exc.errors()}")
            return

        reply_topic = f"/reply/{request.reply}"
        if request.request == DotBotRequestType.DOTBOTS:
            logger.info("Publish dotbots")
            data = [
                dotbot.model_dump(exclude_none=True)
                for dotbot in self.get_dotbots(DotBotQueryModel())
            ]
            message = DotBotReplyModel(
                request=DotBotRequestType.DOTBOTS,
                data=data,
            ).model_dump(exclude_none=True)
            self.qrkey.publish(reply_topic, message)
        elif request.request == DotBotRequestType.LH2_CALIBRATION_STATE:
            logger.info("Publish LH2 state")
            message = DotBotReplyModel(
                request=DotBotRequestType.LH2_CALIBRATION_STATE,
                data=self.lh2_manager.state_model.model_dump(),
            ).model_dump(exclude_none=True)
            self.qrkey.publish(reply_topic, message)
        else:
            logger.warning("Unsupported request command")

    async def _start_serial(self):
        """Starts the serial listener thread in a coroutine."""
        queue = asyncio.Queue()
        event_loop = asyncio.get_event_loop()

        def on_byte_received(byte):
            """Callback called on byte received."""
            event_loop.call_soon_threadsafe(queue.put_nowait, byte)

        async def _wait_for_handshake(queue):
            """Waits for handshake reply and checks it."""
            try:
                byte = await queue.get()
            except asyncio.exceptions.CancelledError as exc:
                raise SerialInterfaceException("Handshake timeout") from exc
            if int.from_bytes(byte, byteorder="little") != PROTOCOL_VERSION:
                raise SerialInterfaceException("Handshake failed")

        if self.settings.port == "sailbot-simulator":
            self.serial = SailBotSimulatorSerialInterface(on_byte_received)
        elif self.settings.port == "dotbot-simulator":
            self.serial = DotBotSimulatorSerialInterface(on_byte_received)
        else:
            self.serial = SerialInterface(
                self.settings.port, self.settings.baudrate, on_byte_received
            )
            self.serial.write(
                int(PROTOCOL_VERSION).to_bytes(
                    length=1, byteorder="little", signed=False
                )
            )
            if self.settings.handshake is True:
                await asyncio.wait_for(_wait_for_handshake(queue), timeout=0.2)
                self.logger.info("Serial handshake success")

        while 1:
            byte = await queue.get()
            self.handle_byte(byte)

    async def _open_webbrowser(self):
        """Wait until the server is ready before opening a web browser."""
        while 1:
            try:
                _, writer = await asyncio.open_connection(
                    "127.0.0.1", self.settings.controller_port
                )
            except ConnectionRefusedError:
                await asyncio.sleep(0.1)
            else:
                writer.close()
                break
        if self.settings.webbrowser is True:
            url = f"http://localhost:{self.settings.controller_port}/PyDotBot?pin={self.qrkey.pin_code}"
            self.logger.info("Opening webbrowser", url=url)
            webbrowser.open(url)

    async def _dotbots_status_refresh(self):
        """Coroutine that periodically updates the status of known dotbot."""
        while 1:
            needs_refresh = [False] * len(self.dotbots)
            for idx, dotbot in enumerate(self.dotbots.values()):
                previous_status = dotbot.status
                if dotbot.last_seen + DEAD_DELAY < time.time():
                    dotbot.status = DotBotStatus.DEAD
                elif dotbot.last_seen + LOST_DELAY < time.time():
                    dotbot.status = DotBotStatus.LOST
                else:
                    dotbot.status = DotBotStatus.ALIVE
                logger = self.logger.bind(
                    source=dotbot.address,
                    application=dotbot.application.name,
                )
                if len(needs_refresh) > idx:
                    needs_refresh[idx] = bool(previous_status != dotbot.status)
                    if needs_refresh[idx]:
                        logger.info(
                            "Dotbot status changed",
                            previous_status=previous_status.name,
                            status=dotbot.status.name,
                        )
            if any(needs_refresh) is True:
                await self.notify_clients(
                    DotBotNotificationModel(cmd=DotBotNotificationCommand.RELOAD)
                )
            await asyncio.sleep(1)

    def _compute_lh2_position(
        self, payload: ProtocolPayload
    ) -> Optional[DotBotLH2Position]:
        if payload.payload_type not in (
            PayloadType.LH2_RAW_DATA,
            PayloadType.DOTBOT_DATA,
        ):
            return None
        self.lh2_manager.last_raw_data = payload.values
        if self.lh2_manager.state != LighthouseManagerState.Calibrated:
            return None
        return self.lh2_manager.compute_position(payload.values)

    def handle_byte(self, byte):
        """Called on each byte received over UART."""
        self.hdlc_handler.handle_byte(byte)
        if self.hdlc_handler.state == HDLCState.READY:
            payload = self.hdlc_handler.payload
            if payload:
                try:
                    payload = ProtocolPayload.from_bytes(payload)
                except ProtocolPayloadParserException:
                    self.logger.warning("Cannot parse payload")
                    if self.settings.verbose is True:
                        print(payload)
                    return
                self.handle_received_payload(payload)

    def handle_received_payload(
        self, payload: ProtocolPayload
    ):  # pylint:disable=too-many-branches,too-many-statements
        """Handle a received payload."""
        # Controller is not interested by command messages received
        if payload.payload_type in [
            PayloadType.CMD_MOVE_RAW,
            PayloadType.CMD_RGB_LED,
        ]:
            return
        source = hexlify(int(payload.header.source).to_bytes(8, "big")).decode()
        logger = self.logger.bind(
            source=source,
            payload_type=payload.payload_type.name,
            application=payload.header.application.name,
            msg_id=payload.header.msg_id,
        )
        if source == GATEWAY_ADDRESS_DEFAULT:
            logger.warning("Invalid source in payload")
            return
        dotbot = DotBotModel(
            address=source,
            application=payload.header.application,
            last_seen=time.time(),
        )
        notification_cmd = DotBotNotificationCommand.NONE
        if source in self.dotbots:
            dotbot.mode = self.dotbots[source].mode
            dotbot.status = self.dotbots[source].status
            dotbot.direction = self.dotbots[source].direction
            dotbot.wind_angle = self.dotbots[source].wind_angle
            dotbot.rudder_angle = self.dotbots[source].rudder_angle
            dotbot.sail_angle = self.dotbots[source].sail_angle
            dotbot.rgb_led = self.dotbots[source].rgb_led
            dotbot.lh2_position = self.dotbots[source].lh2_position
            dotbot.gps_position = self.dotbots[source].gps_position
            dotbot.waypoints = self.dotbots[source].waypoints
            dotbot.waypoints_threshold = self.dotbots[source].waypoints_threshold
            dotbot.position_history = self.dotbots[source].position_history
        else:
            # reload if a new dotbot comes in
            logger.info("New dotbot")
            notification_cmd = DotBotNotificationCommand.RELOAD

        if (
            payload.payload_type in [PayloadType.DOTBOT_DATA, PayloadType.SAILBOT_DATA]
            and -500 <= payload.values.direction <= 500
        ):
            dotbot.direction = payload.values.direction
            logger = logger.bind(direction=dotbot.direction)

        if payload.payload_type in [PayloadType.SAILBOT_DATA]:
            logger = logger.bind(
                wind_angle=dotbot.wind_angle,
                rudder_angle=dotbot.rudder_angle,
                sail_angle=dotbot.sail_angle,
            )

        dotbot.lh2_position = self._compute_lh2_position(payload)
        if (
            dotbot.lh2_position is not None
            and 0 <= dotbot.lh2_position.x <= 1
            and 0 <= dotbot.lh2_position.y <= 1
        ):
            new_position = DotBotLH2Position(
                x=dotbot.lh2_position.x,
                y=dotbot.lh2_position.y,
                z=dotbot.lh2_position.z,
            )
            logger.info("lh2", x=dotbot.lh2_position.x, y=dotbot.lh2_position.y)
            if (
                not dotbot.position_history
                or lh2_distance(dotbot.position_history[-1], new_position)
                >= LH2_POSITION_DISTANCE_THRESHOLD
            ):
                dotbot.position_history.append(new_position)
                notification_cmd = DotBotNotificationCommand.UPDATE
            if len(dotbot.position_history) > MAX_POSITION_HISTORY_SIZE:
                dotbot.position_history.pop(0)
            # Send the computed position back to the dotbot
            header = ProtocolHeader(
                destination=int(source, 16),
                source=int(self.settings.gw_address, 16),
                swarm_id=int(self.settings.swarm_id, 16),
                application=dotbot.application,
                version=PROTOCOL_VERSION,
            )
            self.send_payload(
                ProtocolPayload(
                    header,
                    PayloadType.LH2_LOCATION,
                    LH2Location(
                        int(dotbot.lh2_position.x * 1e6),
                        int(dotbot.lh2_position.y * 1e6),
                        int(dotbot.lh2_position.z * 1e6),
                    ),
                )
            )
        elif payload.payload_type == PayloadType.DOTBOT_DATA:
            logger.warning("lh2: invalid position")

        if payload.payload_type == PayloadType.DOTBOT_SIMULATOR_DATA:
            dotbot.direction = payload.values.theta
            new_position = DotBotLH2Position(
                x=payload.values.pos_x / 1e6,
                y=payload.values.pos_y / 1e6,
                z=0,
            )
            dotbot.lh2_position = new_position
            dotbot.position_history.append(new_position)
            notification_cmd = DotBotNotificationCommand.UPDATE

        if payload.payload_type in [PayloadType.GPS_POSITION, PayloadType.SAILBOT_DATA]:
            new_position = DotBotGPSPosition(
                latitude=float(payload.values.latitude) / 1e6,
                longitude=float(payload.values.longitude) / 1e6,
            )
            dotbot.gps_position = new_position
            # Read wind sensor measurements
            dotbot.wind_angle = payload.values.wind_angle
            dotbot.rudder_angle = payload.values.rudder_angle
            dotbot.sail_angle = payload.values.sail_angle
            logger.info(
                "gps",
                lat=new_position.latitude,
                long=new_position.longitude,
                wind_angle=dotbot.wind_angle,
                rudder_angle=dotbot.rudder_angle,
                sail_angle=dotbot.sail_angle,
            )
            if (
                not dotbot.position_history
                or gps_distance(dotbot.position_history[-1], new_position)
                >= GPS_POSITION_DISTANCE_THRESHOLD
            ):
                dotbot.position_history.append(new_position)
            if len(dotbot.position_history) > MAX_POSITION_HISTORY_SIZE:
                dotbot.position_history.pop(0)
            notification_cmd = DotBotNotificationCommand.UPDATE

        if notification_cmd == DotBotNotificationCommand.UPDATE:
            notification = DotBotNotificationModel(
                cmd=notification_cmd.value,
                data=DotBotNotificationUpdate(
                    address=dotbot.address,
                    direction=dotbot.direction,
                    wind_angle=dotbot.wind_angle,
                    rudder_angle=dotbot.rudder_angle,
                    sail_angle=dotbot.sail_angle,
                    lh2_position=dotbot.lh2_position,
                    gps_position=dotbot.gps_position,
                ),
            )
        else:
            notification = DotBotNotificationModel(cmd=notification_cmd.value)

        if self.settings.verbose is True:
            print(payload)
        self.dotbots.update({dotbot.address: dotbot})
        if notification_cmd != DotBotNotificationCommand.NONE:
            asyncio.create_task(self.notify_clients(notification))

    async def _ws_send_safe(self, websocket: WebSocket, msg: str):
        """Safely send a message to a websocket client."""
        try:
            await websocket.send_text(msg)
        except websockets.exceptions.ConnectionClosedError:
            await asyncio.sleep(0.1)

    async def notify_clients(self, notification):
        """Send a message to all clients connected."""
        self.logger.debug("notify", cmd=notification.cmd.name)
        await asyncio.gather(
            *[
                self._ws_send_safe(
                    websocket, json.dumps(notification.model_dump(exclude_none=True))
                )
                for websocket in self.websockets
            ]
        )
        self.qrkey.publish("/notify", notification.model_dump(exclude_none=True))

    def send_payload(self, payload: ProtocolPayload):
        """Sends a command in an HDLC frame over serial."""
        destination = hexlify(
            int(payload.header.destination).to_bytes(8, "big")
        ).decode()
        if destination not in self.dotbots:
            return
        # make sure the application in the payload matches the bot application
        payload.header.application = self.dotbots[destination].application
        if self.serial is not None:
            self.serial.write(hdlc_encode(payload.to_bytes()))
            self.logger.debug(
                "Payload sent",
                application=payload.header.application.name,
                destination=destination,
                payload_type=payload.payload_type.name,
            )

    def get_dotbots(self, query: DotBotQueryModel) -> List[DotBotModel]:
        """Returns the list of dotbots matching the query."""
        dotbots: List[DotBotModel] = []
        for dotbot in self.dotbots.values():
            if (
                query.application is not None
                and dotbot.application != query.application
            ):
                continue
            if query.mode is not None and dotbot.mode != query.mode:
                continue
            if query.status is not None and dotbot.status != query.status:
                continue
            _dotbot = DotBotModel(**dotbot.model_dump())
            _dotbot.position_history = _dotbot.position_history[: query.max_positions]
            dotbots.append(_dotbot)
        return sorted(dotbots, key=lambda dotbot: dotbot.address)

    async def web(self):
        """Starts the web server application."""
        logger = LOGGER.bind(context=__name__)
        config = uvicorn.Config(
            api, port=self.settings.controller_port, log_level="critical"
        )
        server = uvicorn.Server(config)

        try:
            logger.info("Starting web server")
            await server.serve()
        except asyncio.exceptions.CancelledError:
            logger.info("Web server cancelled")
        else:
            logger.info("Stopping web server")
            raise SystemExit()

    async def run(self):
        """Launch the controller."""
        tasks = []
        try:
            tasks = [
                asyncio.create_task(self.qrkey.start(subscriptions=self.subscriptions)),
                asyncio.create_task(self.web()),
                asyncio.create_task(self._open_webbrowser()),
                asyncio.create_task(self._start_serial()),
                asyncio.create_task(self._dotbots_status_refresh()),
            ]
            await asyncio.gather(*tasks)
        except (
            SerialInterfaceException,
            serial.serialutil.SerialException,
        ) as exc:
            self.logger.error(f"Error: {exc}")
        except SystemExit:
            self.logger.info("Stopping controller")
        finally:
            for task in tasks:
                task.cancel()



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for the web server application."""

import os
from typing import List

from fastapi import Depends, FastAPI, HTTPException, WebSocket, WebSocketDisconnect
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles

from dotbot import pydotbot_version
from dotbot.models import (
    DotBotCalibrationStateModel,
    DotBotModel,
    DotBotMoveRawCommandModel,
    DotBotNotificationCommand,
    DotBotNotificationModel,
    DotBotQueryModel,
    DotBotRgbLedCommandModel,
    DotBotWaypoints,
)
from dotbot.protocol import (
    PROTOCOL_VERSION,
    ApplicationType,
    CommandMoveRaw,
    CommandRgbLed,
    GPSPosition,
    GPSWaypoints,
    LH2Location,
    LH2Waypoints,
    PayloadType,
    ProtocolHeader,
    ProtocolPayload,
)

PYDOTBOT_FRONTEND_BASE_URL = os.getenv(
    "PYDOTBOT_FRONTEND_BASE_URL", "https://dotbots.github.io/PyDotBot"
)


api = FastAPI(
    debug=0,
    title="DotBot controller API",
    description="This is the DotBot controller API",
    version=pydotbot_version(),
    docs_url="/api",
    redoc_url=None,
)
api.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@api.put(
    path="/controller/dotbots/{address}/{application}/move_raw",
    summary="Move the dotbot",
    tags=["dotbots"],
)
async def dotbots_move_raw(
    address: str, application: int, command: DotBotMoveRawCommandModel
):
    """Set the current active DotBot."""
    if address not in api.controller.dotbots:
        raise HTTPException(status_code=404, detail="No matching dotbot found")

    header = ProtocolHeader(
        destination=int(address, 16),
        source=int(api.controller.settings.gw_address, 16),
        swarm_id=int(api.controller.settings.swarm_id, 16),
        application=ApplicationType(application),
        version=PROTOCOL_VERSION,
    )
    payload = ProtocolPayload(
        header,
        PayloadType.CMD_MOVE_RAW,
        CommandMoveRaw(
            command.left_x, command.left_y, command.right_x, command.right_y
        ),
    )
    api.controller.send_payload(payload)
    api.controller.dotbots[address].move_raw = command


@api.put(
    path="/controller/dotbots/{address}/{application}/rgb_led",
    summary="Set the dotbot RGB LED color",
    tags=["dotbots"],
)
async def dotbots_rgb_led(
    address: str, application: int, command: DotBotRgbLedCommandModel
):
    """Set the current active DotBot."""
    if address not in api.controller.dotbots:
        raise HTTPException(status_code=404, detail="No matching dotbot found")

    header = ProtocolHeader(
        destination=int(address, 16),
        source=int(api.controller.settings.gw_address, 16),
        swarm_id=int(api.controller.settings.swarm_id, 16),
        application=ApplicationType(application),
        version=PROTOCOL_VERSION,
    )
    payload = ProtocolPayload(
        header,
        PayloadType.CMD_RGB_LED,
        CommandRgbLed(command.red, command.green, command.blue),
    )
    api.controller.send_payload(payload)
    api.controller.dotbots[address].rgb_led = command


@api.put(
    path="/controller/dotbots/{address}/{application}/waypoints",
    summary="Set the dotbot control mode",
    tags=["dotbots"],
)
async def dotbots_waypoints(
    address: str,
    application: int,
    waypoints: DotBotWaypoints,
):
    """Set the waypoints of a DotBot."""
    if address not in api.controller.dotbots:
        raise HTTPException(status_code=404, detail="No matching dotbot found")

    header = ProtocolHeader(
        destination=int(address, 16),
        source=int(api.controller.settings.gw_address, 16),
        swarm_id=int(api.controller.settings.swarm_id, 16),
        application=ApplicationType(application),
        version=PROTOCOL_VERSION,
    )
    waypoints_list = waypoints.waypoints
    if ApplicationType(application) == ApplicationType.SailBot:
        if api.controller.dotbots[address].gps_position is not None:
            waypoints_list = [
                api.controller.dotbots[address].gps_position
            ] + waypoints.waypoints
        payload = ProtocolPayload(
            header,
            PayloadType.GPS_WAYPOINTS,
            GPSWaypoints(
                threshold=waypoints.threshold,
                waypoints=[
                    GPSPosition(
                        latitude=int(waypoint.latitude * 1e6),
                        longitude=int(waypoint.longitude * 1e6),
                    )
                    for waypoint in waypoints.waypoints
                ],
            ),
        )
    else:  # DotBot application
        if api.controller.dotbots[address].lh2_position is not None:
            waypoints_list = [
                api.controller.dotbots[address].lh2_position
            ] + waypoints.waypoints
        payload = ProtocolPayload(
            header,
            PayloadType.LH2_WAYPOINTS,
            LH2Waypoints(
                threshold=waypoints.threshold,
                waypoints=[
                    LH2Location(
                        pos_x=int(waypoint.x * 1e6),
                        pos_y=int(waypoint.y * 1e6),
                        pos_z=int(waypoint.z * 1e6),
                    )
                    for waypoint in waypoints.waypoints
                ],
            ),
        )
    api.controller.dotbots[address].waypoints = waypoints_list
    api.controller.dotbots[address].waypoints_threshold = waypoints.threshold
    api.controller.send_payload(payload)
    await api.controller.notify_clients(
        DotBotNotificationModel(cmd=DotBotNotificationCommand.RELOAD)
    )


@api.delete(
    path="/controller/dotbots/{address}/positions",
    summary="Clear the history of positions of a DotBot",
    tags=["dotbots"],
)
async def dotbot_positions_history_clear(address: str):
    """Clear the history of positions of a dotbot."""
    if address not in api.controller.dotbots:
        raise HTTPException(status_code=404, detail="No matching dotbot found")
    api.controller.dotbots[address].position_history = []


@api.get(
    path="/controller/dotbots/{address}",
    response_model=DotBotModel,
    response_model_exclude_none=True,
    summary="Return information about a dotbot given its address",
    tags=["dotbots"],
)
async def dotbot(address: str, query: DotBotQueryModel = Depends()):
    """Dotbot HTTP GET handler."""
    if address not in api.controller.dotbots:
        raise HTTPException(status_code=404, detail="No matching dotbot found")
    _dotbot = DotBotModel(**api.controller.dotbots[address].model_dump())
    _dotbot.position_history = _dotbot.position_history[: query.max_positions]
    return _dotbot


@api.get(
    path="/controller/dotbots",
    response_model=List[DotBotModel],
    response_model_exclude_none=True,
    summary="Return the list of available dotbots",
    tags=["dotbots"],
)
async def dotbots(query: DotBotQueryModel = Depends()):
    """Dotbots HTTP GET handler."""
    return api.controller.get_dotbots(query)


@api.post(
    path="/controller/lh2/calibration/{point_idx}",
    summary="Trigger the acquisition of one LH2 point",
    tags=["dotbots"],
)
async def controller_add_lh2_calibration_point(point_idx: int):
    """LH2 calibration, add single calibration point."""
    api.controller.lh2_manager.add_calibration_point(point_idx)


@api.put(
    path="/controller/lh2/calibration",
    summary="Trigger a computation of the LH2 calibration",
    tags=["dotbots"],
)
async def controller_apply_lh2_calibration():
    """Apply LH2 calibration."""
    api.controller.lh2_manager.compute_calibration()


@api.get(
    path="/controller/lh2/calibration",
    response_model=DotBotCalibrationStateModel,
    response_model_exclude_none=True,
    summary="Return the LH2 calibration state",
    tags=["dotbots"],
)
async def controller_get_lh2_calibration():
    """LH2 calibration GET handler."""
    return api.controller.lh2_manager.state_model


@api.websocket("/controller/ws/status")
async def websocket_endpoint(websocket: WebSocket):
    """Websocket server endpoint."""
    await websocket.accept()
    api.controller.websockets.append(websocket)
    try:
        while True:
            _ = await websocket.receive_text()
    except WebSocketDisconnect:
        if websocket in api.controller.websockets:
            api.controller.websockets.remove(websocket)


# Mount static files after all routes are defined
FRONTEND_DIR = os.path.join(os.path.dirname(__file__), "frontend", "build")
api.mount("/PyDotBot", StaticFiles(directory=FRONTEND_DIR, html=True), name="PyDotBot")



"""Test module for the main function."""

from importlib.metadata import PackageNotFoundError
from unittest.mock import patch

import serial
from click.testing import CliRunner

from dotbot.main import main
from dotbot.serial_interface import get_default_port

MAIN_HELP_EXPECTED = f"""Usage: main [OPTIONS]

  DotBotController, universal SailBot and DotBot controller.

Options:
  -p, --port TEXT                 Virtual com port. Defaults to '{get_default_port()}'
  -b, --baudrate INTEGER          Serial baudrate. Defaults to 1000000
  -d, --dotbot-address TEXT       Address in hex of the DotBot to control.
                                  Defaults to FFFFFFFFFFFFFFFF
  -g, --gw-address TEXT           Gateway address in hex. Defaults to
                                  0000000000000000
  -s, --swarm-id TEXT             Swarm ID in hex. Defaults to 0000
  -c, --controller-port INTEGER   Controller port. Defaults to '8000'
  -w, --webbrowser                Open a web browser automatically
  -v, --verbose                   Run in verbose mode (all payloads received are
                                  printed in terminal)
  --log-level [debug|info|warning|error]
                                  Logging level. Defaults to info
  --log-output PATH               Filename where logs are redirected
  --handshake                     Perform a basic handshake with the gateway
                                  board on startup
  --help                          Show this message and exit.
"""


def test_main_help():
    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0
    assert result.output == MAIN_HELP_EXPECTED


@patch("dotbot.serial_interface.serial.Serial.open")
@patch("dotbot.controller.QrkeyController")
@patch("dotbot.version")
@patch("dotbot.controller.Controller.run")
def test_main(run, version, _, __):
    version.return_value = "test"
    runner = CliRunner()
    result = runner.invoke(main)
    assert result.exit_code == 0
    assert "Welcome to the DotBots controller (version: test)." in result.output
    run.assert_called_once()

    version.side_effect = PackageNotFoundError
    result = runner.invoke(main)
    assert result.exit_code == 0
    assert "Welcome to the DotBots controller (version: unknown)." in result.output


@patch("dotbot.serial_interface.serial.Serial.open")
@patch("dotbot.controller.QrkeyController")
@patch("dotbot.controller.Controller.run")
def test_main_interrupts(run, _, __):
    runner = CliRunner()
    run.side_effect = KeyboardInterrupt
    result = runner.invoke(main)
    assert result.exit_code == 0

    runner = CliRunner()
    run.side_effect = SystemExit
    result = runner.invoke(main)
    assert result.exit_code == 0

    run.side_effect = serial.serialutil.SerialException("serial test error")
    result = runner.invoke(main)
    assert result.exit_code != 0
    assert "Serial error: serial test error" in result.output



"""Pydotbot module."""

from importlib.metadata import PackageNotFoundError, version

from dotbot.serial_interface import get_default_port

SERIAL_PORT_DEFAULT = get_default_port()
SERIAL_BAUDRATE_DEFAULT = 1000000
DOTBOT_ADDRESS_DEFAULT = "FFFFFFFFFFFFFFFF"  # Broadcast by default
GATEWAY_ADDRESS_DEFAULT = "0000000000000000"
SWARM_ID_DEFAULT = "0000"
CONTROLLER_PROTOCOL_DEFAULT = "http"
CONTROLLER_HOSTNAME_DEFAULT = "localhost"
CONTROLLER_PORT_DEFAULT = "8000"


def pydotbot_version() -> str:
    """Returns the version of the pydotdot package."""
    try:
        return version("pydotbot")
    except PackageNotFoundError:
        return "unknown"



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Filip Maksimovic <filip.maksimovic@inria.fr>
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module containing the API to convert LH2 raw data to relative positions."""

# pylint: disable=invalid-name,unspecified-encoding,no-member

import math
import os
import pickle
import sys
from ctypes import CDLL
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, Optional

import cv2
import numpy as np

from dotbot.logger import LOGGER
from dotbot.models import DotBotCalibrationStateModel, DotBotLH2Position
from dotbot.protocol import Lh2RawData

if sys.platform == "win32":
    LIB_EXT = "dll"
elif sys.platform == "darwin":
    LIB_EXT = "dylib"
else:
    LIB_EXT = "so"

LH2_LIB_PATH = os.path.join(os.path.dirname(__file__), "lib", f"lh2.{LIB_EXT}")
LH2_LIB = CDLL(LH2_LIB_PATH)
REFERENCE_POINTS_DEFAULT = [
    [-0.1, 0.1],
    [0.1, 0.1],
    [-0.1, -0.1],
    [0.1, -0.1],
]
CALIBRATION_DIR = Path.home() / ".pydotbot"


def _lh2_raw_data_to_counts(raw_data: Lh2RawData, func: callable) -> List[int]:
    counts = [0] * 2
    pos_A = 0
    pos_B = 0
    for i in range(2):
        index = 0
        if raw_data.locations[i].polynomial_index in [0, 1]:
            index = pos_A
            pos_A += 1
        elif raw_data.locations[i].polynomial_index in [0, 1]:
            index = 2 + pos_B
            pos_B += 1
        if index > 1:
            continue
        counts[index] = func(
            raw_data.locations[i].polynomial_index,
            raw_data.locations[i].bits >> (47 - raw_data.locations[i].offset),
        )
    return counts


def lh2_raw_data_to_counts(raw_data: Lh2RawData) -> List[int]:
    """Convert bits sequence to an array of counts."""
    return _lh2_raw_data_to_counts(raw_data, LH2_LIB.reverse_count_p)


def calculate_camera_point(count1, count2, poly_in):
    """Calculate camera points from counts."""
    if poly_in < 2:
        period = 959000
    if poly_in > 1:
        period = 957000

    a1 = (count1 * 8 / period) * 2 * math.pi
    a2 = (count2 * 8 / period) * 2 * math.pi

    cam_x = -math.tan(0.5 * (a1 + a2))
    if count1 < count2:
        cam_y = -math.sin(a2 / 2 - a1 / 2 - 60 * math.pi / 180) / math.tan(math.pi / 6)
    else:
        cam_y = -math.sin(a1 / 2 - a2 / 2 - 60 * math.pi / 180) / math.tan(math.pi / 6)

    return cam_x, cam_y


def _unitize(x_in, y_in):
    magnitude = np.sqrt(x_in**2 + y_in**2)
    return x_in / magnitude, y_in / magnitude


@dataclass
class CalibrationData:
    """Class that stores calibration data."""

    zeta: float
    random_rodriguez: np.array
    normal: np.array
    m: np.array


class LighthouseManagerState(Enum):
    """Enum for lighthouse manager internal state."""

    NotCalibrated = 0
    CalibrationInProgress = 1
    Ready = 2
    Calibrated = 3


class LighthouseManager:
    """Class to manage the LightHouse positionning state and workflow."""

    def __init__(self):
        self.state = LighthouseManagerState.NotCalibrated
        self.reference_points = REFERENCE_POINTS_DEFAULT
        Path.mkdir(CALIBRATION_DIR, exist_ok=True)
        self.calibration_output_path = CALIBRATION_DIR / "calibration.out"
        self.calibration_data = self._load_calibration()
        self.calibration_points = np.zeros(
            (2, len(self.reference_points), 2), dtype=np.float64
        )
        self.calibration_points_available = [False] * len(self.reference_points)
        self.last_raw_data = None
        self.logger = LOGGER.bind(context=__name__)
        self.logger.info("Lighthouse initialized")

    @property
    def state_model(self) -> DotBotCalibrationStateModel:
        """Return the state as pydantic model."""
        if self.state == LighthouseManagerState.CalibrationInProgress:
            return DotBotCalibrationStateModel(state="running")
        if self.state == LighthouseManagerState.Ready:
            return DotBotCalibrationStateModel(state="ready")
        if self.state == LighthouseManagerState.Calibrated:
            return DotBotCalibrationStateModel(state="done")
        return DotBotCalibrationStateModel(state="unknown")

    def _load_calibration(self) -> Optional[CalibrationData]:
        if not os.path.exists(self.calibration_output_path):
            return None
        with open(self.calibration_output_path, "rb") as calibration_file:
            calibration = pickle.load(calibration_file)
        self.state = LighthouseManagerState.Calibrated
        return calibration

    def add_calibration_point(self, index):
        """Register a new camera points for calibration."""
        if self.last_raw_data is None:
            self.logger.warning("Missing raw data", index=index)
            return

        self.calibration_points_available[index] = True

        counts = lh2_raw_data_to_counts(self.last_raw_data)
        self.calibration_points[0][index] = np.asarray(
            calculate_camera_point(
                counts[0],
                counts[1],
                self.last_raw_data.locations[0].polynomial_index,
            ),
            dtype=np.float64,
        )
        self.calibration_points[1][index] = np.asarray(
            calculate_camera_point(
                counts[0],
                counts[1],
                self.last_raw_data.locations[1].polynomial_index,
            ),
            dtype=np.float64,
        )

        if all(self.calibration_points_available) is False:
            self.state = LighthouseManagerState.CalibrationInProgress
        if all(self.calibration_points_available) is True:
            self.state = LighthouseManagerState.Ready
        self.logger.info("Calibration point added", index=index, state=self.state)

    def compute_calibration(self):  # pylint: disable=too-many-locals
        """Compute the calibration values and matrices."""
        if self.state != LighthouseManagerState.Ready:
            self.logger.warning("Not ready, skipping calibration")
            return

        self.logger.info("Calibrating", points=self.calibration_points)

        camera_points = [[], []]
        for data in self.calibration_points[0]:
            camera_points[0].append(data)
        for data in self.calibration_points[1]:
            camera_points[1].append(data)
        camera_points_arr = np.asarray(camera_points, dtype=np.float64)
        homography_mat = cv2.findHomography(
            camera_points_arr[0][0 : len(camera_points[0])][:],
            camera_points_arr[1][0 : len(camera_points[1])][:],
            method=cv2.RANSAC,
            ransacReprojThreshold=0.001,
        )[0]

        _, S, V = np.linalg.svd(homography_mat)
        V = V.T

        s1 = S[0] / S[1]
        s3 = S[2] / S[1]
        zeta = s1 - s3
        a1 = np.sqrt(1 - s3**2)
        b1 = np.sqrt(s1**2 - 1)
        a, b = _unitize(a1, b1)
        v1 = np.array(V[:, 0])
        v3 = np.array(V[:, 2])
        n = b * v1 + a * v3

        if n[2] < 0:
            n = -n

        random_rodriguez = np.array(
            [
                [
                    -n[1] / np.sqrt(n[0] * n[0] + n[1] * n[1]),
                    n[0] / np.sqrt(n[0] * n[0] + n[1] * n[1]),
                    0,
                ],
                [
                    n[0] * n[2] / np.sqrt(n[0] * n[0] + n[1] * n[1]),
                    n[1] * n[2] / np.sqrt(n[0] * n[0] + n[1] * n[1]),
                    -np.sqrt(n[0] * n[0] + n[1] * n[1]),
                ],
                [-n[0], -n[1], -n[2]],
            ]
        )

        pts_cam_new = np.hstack(
            (camera_points_arr[1], np.ones((len(camera_points_arr[1]), 1)))
        )
        scales = (1 / zeta) / np.matmul(n, pts_cam_new.T)
        scales_matrix = np.vstack((scales, scales, scales))
        final_points = scales_matrix * pts_cam_new.T
        final_points = final_points.T

        M, _ = cv2.findHomography(
            final_points.dot(random_rodriguez.T)[:, 0:2],
            np.array([self.reference_points], dtype=np.float64) + 0.5,
            cv2.RANSAC,
            5.0,
        )

        self.calibration_data = CalibrationData(zeta, random_rodriguez, n, M)

        with open(self.calibration_output_path, "wb") as output_file:
            pickle.dump(self.calibration_data, output_file)

        self.state = LighthouseManagerState.Calibrated
        self.logger.info("Calibration done", data=self.calibration_data)

    def compute_position(self, raw_data: Lh2RawData) -> Optional[DotBotLH2Position]:
        """Compute the position coordinates from LH2 raw data and available calibration."""
        if self.state != LighthouseManagerState.Calibrated:
            return None

        if any(raw_data.locations[index].bits == 0 for index in range(2)):
            return None

        counts = lh2_raw_data_to_counts(raw_data)
        camera_points = np.asarray(
            [
                calculate_camera_point(
                    counts[0], counts[1], raw_data.locations[0].polynomial_index
                ),
                calculate_camera_point(
                    counts[0], counts[1], raw_data.locations[1].polynomial_index
                ),
            ],
            dtype=np.float64,
        )

        pts_cam_new = np.hstack((camera_points, np.ones((len(camera_points), 1))))
        scales = (1 / self.calibration_data.zeta) / np.matmul(
            self.calibration_data.normal, pts_cam_new.T
        )
        scales_matrix = np.vstack((scales, scales, scales))
        final_points = scales_matrix * pts_cam_new.T
        final_points = final_points.T
        corners_planar = final_points.dot(self.calibration_data.random_rodriguez.T)[
            :, 0:2
        ][1].reshape(1, 1, 2)
        pts_meter_corners = cv2.perspectiveTransform(
            corners_planar, self.calibration_data.m
        ).reshape(-1, 2)
        return DotBotLH2Position(
            x=pts_meter_corners[0][0], y=1 - pts_meter_corners[0][1], z=0.0
        )
